ShinyEnv:Use("ShinyStats")

local L = ShinyLocale

ShinyStats = LibStub("AceAddon-3.0"):NewAddon("ShinyStats", "AceConsole-3.0")
_G["ShinyStats"] = ShinyStats


local _print = _G["print"]

--[===[@debug@
local print = function(...)
	local frame = ChatFrame5 or DEFAULT_CHAT_FRAME
	frame:AddMessage(strjoin(" ", tostringall(...)))
end
--@end-debug@]===]
--@non-debug@
local print = function(...) return true end
--@end-non-debug@


function ShinyStats:OnInitialize()
	self.playerClass = nil
	self.playerSpecID = nil
	self.playerSpecIndex = nil
	self.tainted = false
	
	self.usefulStats = {}
	self.primaryStat = nil

	self.colors = nil
	self.patterns = nil

	self.tooltipsHooked = false
	self.eventsHooked = false

	ShinyDB:OnInitialize()
	ShinyCache:OnInitialize()
	ShinyConfig:OnInitialize()

	self:CreatePatterns()
end


function ShinyStats:OnEnable()
	self:UpdateClass()
	self:UpdateSpec()

	self:CreateColors()
	
	self:RegisterChatCommand("shiny", "ChatCommand")

	self:HookEvents()
	self:HookTooltips()

	self:UpdateIfTainted()
end

	
function ShinyStats:CreateColors()
	local colors = self.colors or {}
	local color

	for k,data in pairs(ShinyDB.db.profile.colors) do
		color = Color:CreateFromHEXA(data.value)
		color:SetEnabled(data.enabled)

		colors[k] = color
	end

	colors[SHINY_CLASS] = Color:CreateFromAHEX(RAID_CLASS_COLORS[self.playerClass].colorStr)

	self.colors = colors
end

	
function ShinyStats:CreatePatterns()
	local patterns = {}

	for k,v in pairs(PATTERN_CATEGORY_MAP) do
		patterns[k] = {}
	end

	for _, pattern in pairs(PATTERN_MAP) do
		for categoryName, categoryBit in pairs(PATTERN_CATEGORY_MAP) do
			if (pattern:HasCategory(categoryBit)) then
				table.insert(patterns[categoryName], pattern)
			end
		end
	end

	self.patterns = patterns
end


function ShinyStats:ReloadConfig()
	self.tainted = true

	self:CreateColors()
	self:UpdateIfTainted()
end


function ShinyStats:UpdateIfTainted()
	if (not self.tainted) then return end

	self:UpdateUsefulStats()
	self:UpdateDynamicColors()

	ShinyCache:Clear()
	self.tainted = false
end


function ShinyStats:UpdateClass()
	local _, class = UnitClass("player")

	self.playerClass = class
end


function ShinyStats:UpdateSpec()
	--[===[ --[===[@debug@
	-- print("ShinyStats", "Updating Spec!")
	-- --@end-debug@]===]

	local specIndex = GetSpecialization()
	local lastPrimaryStat = self.primaryStat
	local lastSpecID = self.playerSpecID

	if (specIndex) then
		self.playerSpecIndex = specIndex
		self.playerSpecID = CLASS_SPEC_ID_MAP[self.playerClass][specIndex]
		self.primaryStat = select(7, GetSpecializationInfo(specIndex))
		-- local primaryStat = select(7, GetSpecializationInfo(spec, nil, nil, nil, UnitSex("player")))
		-- GetSpecializationInfo(GetSpecialization());
	end

	if (not (specIndex 
		and lastPrimaryStat == self.primaryStat
		and lastSpecID == self.playerSpecID)) then

		self.tainted = true
	end
end


function ShinyStats:UpdateUsefulStats()
	local list = self.usefulStats

	for k,v in pairs(STAT_CLASS_MAP) do
		list[k] = self:IsStatUseful(k, self.playerClass, self.playerSpecIndex) or false
	end
end


function ShinyStats:IsStatUseful(stat, class, spec)
	local stat = STAT_CLASS_MAP[stat][class]

	if (spec and type(stat) == "table") then
		return stat[spec]
	else
		return stat
	end
end


function ShinyStats:UpdateDynamicColors()
	self.colors[SHINY_PRIMARYSTAT] = self.colors[PRIMARY_STAT_MAP[self.primaryStat]]
end


function ShinyStats:ChatCommand(userInput)
	local userInput = userInput:lower()

	if (userInput == "dbg" or userInput == "debug") then
		ShinyDB.db.profile.options.DEBUG = not ShinyDB.db.profile.options.DEBUG
	elseif (userInput == "clearcache" or userInput == "clear") then
		ShinyCache:Clear()
	elseif (userInput == "import") then
		ShinyConfig:ImportConfig()
		ShinyCache:SetTainted(true)
		ShinyStats:CreateColors()
	else
		ShinyConfig:Show()
	end
end


function ShinyStats:IsRecipe(itemLink)
	return (itemLink:match(L["Recipe"]) 
		or itemLink:match(L["Pattern"]) 
		or itemLink:match(L["Plans"]) 
		or itemLink:match(L["Formula"]) 
		or itemLink:match(L["Technique"]) 
		or itemLink:match(L["Design"]) 
		or itemLink:match(L["Manual"]))
end


function ShinyStats:Escape(text)
	return text:gsub("|", "||")
end


function ShinyStats:Strip(text)
	text = text:gsub("|c%x%x%x%x%x%x%x%x", "")
	text = text:gsub("|r", "")

	return text
end


function ShinyStats:Shorten(text)
	local separator = ShinyDB:GetOption("SHORTEN_SEPARATOR")

	return text:gsub("|r" .. L["%sand%s"] .. "|c", "|r" .. separator .. " |c")
end


function ShinyStats:SplitItemLink(itemLink)
	-- item:itemId:enchantId:gemId1:gemId2:gemId3:gemId4:suffixId:uniqueId:linkLevel:specializationID:upgradeId:instanceDifficultyId:numBonusIds:bonusId1:bonusId2:upgradeValue
	-- itemId, enchantId, gemId1, gemId2, gemId3, gemId4, suffixId, uniqueId, linkLevel, specializationID, upgradeId, instanceDifficultyId, numBonusIds, bonusId1, bonusId2, upgradeValue
	return itemLink:match("item:(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*)")
end


function ShinyStats:ValidateSplitItemLink(itemLink)
	-- item:itemId:enchantId:gemId1:gemId2:gemId3:gemId4:suffixId:uniqueId:linkLevel:specializationID:upgradeId:instanceDifficultyId:numBonusIds:bonusId1:bonusId2:upgradeValue
	-- itemId, enchantId, gemId1, gemId2, gemId3, gemId4, suffixId, uniqueId, linkLevel, specializationID, upgradeId, instanceDifficultyId, numBonusIds, bonusId1, bonusId2, upgradeValue
	local result = {}

	for i,v in ipairs{self:SplitItemLink(itemLink)} do
		if (v and v ~= "") then
			result[i] = v
		end
	end
	
	return unpack(result)
end


function ShinyStats:ShouldSkip(region)
	local text = region:GetText()
	local r,g,b,a = region:GetTextColor()

	return (text:match(L["^Socket Bonus"]) and r < 0.51 and g < 0.51)				-- Inactive Socket Bonus
		or (text:match(L["^%(*%d*%)*%s*Set"]) and r < 0.51 and g < 0.51)			-- Inactive Set Bonus
		or text:match(L["^Legacy Set"])												-- Legacy Set
		or text:match(L["^<Shift"]) 												-- Shift-Click to ...
		or text:match(L["^<Made"]) 													-- Made by ...
		or text:match(L["^<Right"]) 												-- Right-Click to ...
		or text:match(L["^Season"]) 												-- Season (PvP)
		or text:match(L["^Requires"]) 												-- Requires ...
		or text:match(L["^Durability"]) 											-- Durability
		or text:match(L["^Reforged"])												-- Deprecated (Removed)
end


function ShinyStats:ParseRelic(text)
	local found, newColor = true

	if (CLASS_RELIC_MAP[self.playerClass][self.playerSpecIndex][text]) then
		newColor = SHINY_ARTIFACT_RELIC_ACTIVE

	elseif (CLASS_RELIC_MAP[self.playerClass][text]) then
		newColor = SHINY_ARTIFACT_RELIC_USABLE

	else
		newColor = SHINY_ARTIFACT_RELIC_UNUSABLE
	end

	return found, newColor
end


function ShinyStats:ParseEquip(text)
	local found, newText, newColor = false
	
	found, newText, newColor = self:ParseEquipTransform(text)

	if (not found) then
		found, newText, newColor = self:ParseEquipSimplify(text)
	end

	return found, newText, newColor
end


function ShinyStats:ParseEquipTransform(text)
	local found, newText, newColor

	found, newText, newColor = self:ParseEquipProc(text)
	
	if (not found) then
		found, newText, newColor = self:ParseEquipEffect(text)
	end

	return found, newText, newColor
end


function ShinyStats:ParseEquipProc(text)
	local found, newText, newColor = false

	local procTitle = self.colors[SHINY_EQUIP_PROC_TITLE]

	for _,data in ipairs(self.patterns[PATTERN_EQUIP_PROC]) do
		local pattern, colorKey = data:Unpack()
		local _, _, proc = string.find(text, pattern)

		if (proc) then
			newText = "|c" .. procTitle:ToAHEX() .. EQUIP_PROC_TRANSFORMATION_MAP[pattern] .. "|r\n" ..
					  proc:gsub("^%l", string.upper)

			found = true
			newColor = colorKey
			
			break
		end
	end

	return found, newText, newColor
end


function ShinyStats:ParseEquipEffect(text)
	local found, newText, newColor = false

	local effectTitle = self.colors[SHINY_EQUIP_EFFECT_TITLE]

	for _,data in ipairs(self.patterns[PATTERN_EQUIP_EFFECT]) do
		local pattern, colorKey = data:Unpack()
		local _, _, effect = string.find(text, pattern)

		if (effect) then
			newText = "|c" .. effectTitle:ToAHEX() .. EQUIP_EFFECT_TRANSFORMATION_MAP[pattern] .. "|r\n" ..
					  effect:gsub("^%l", string.upper)

			found = true
			newColor = colorKey
			
			break
		end
	end

	return found, newText, newColor
end


function ShinyStats:ParseEquipSimplify(text)
	local found, newText, newColor = false

	return found, newText, newColor
end


function ShinyStats:ParseUse(text)
	local found, newText, newColor = false

	if (ShinyDB:GetOption("ENABLE_USE_TRANSFORMATION")) then
		found, newText, newColor = self:ParseUseTransform(text)
	end

	if (not found and ShinyDB:GetOption("ENABLE_USE_SIMPLIFY")) then
		found, newText, newColor = self:ParseUseSimplify(text)
	end

	if (ShinyDB:GetOption("SHORTEN_USE")) then
		newText = self:Shorten(newText or text)
	end

	if (ShinyDB:GetOption("ENABLE_PARTIAL_MATCHES")) then
		partialFind, newText = self:ParsePartial(newText or text, PATTERN_PARTIAL)
	end

	found = found or partialFind

	return found, newText, newColor
end


function ShinyStats:ParseUseTransform(text)
	local found, newText, newColor = false

	local titleColor = (self.colors[SHINY_USE_TRANSFORM_TITLE]):ToAHEX()
	local descColor = (self.colors[SHINY_USE_TRANSFORM]):ToAHEX()
	local cdColor = (self.colors[SHINY_USE_TRANSFORM_CD]):ToAHEX()

	local descColorKey = SHINY_USE_TRANSFORM

	local title, desc

	local typeStringMap = {
		[TYPE_HEALTH] = L["health"],
		[TYPE_MANA] = L["mana"],
	}

	local replacements = {
		[L["({STAT_NUMBER}) [Hh]ealth"]] 					= { SHINY_RESTORE_HEALTH, TYPE_HEALTH },
		[L["({STAT_NUMBER}) to ({STAT_NUMBER}) [Hh]ealth"]] = { SHINY_RESTORE_HEALTH, TYPE_HEALTH },
		[L["({STAT_NUMBER}) [Mm]ana"]]   					= { SHINY_RESTORE_MANA, TYPE_MANA },
		[L["({STAT_NUMBER}) to ({STAT_NUMBER}) [Mm]ana"]]   = { SHINY_RESTORE_MANA, TYPE_MANA },
	}

	local feastPatterns = {
		L["[Ff]eed up to %d+ people in your raid or party"],
		L["[Ff]eed your entire raid or party"],
		L["[Ff]eed your party"],
	}

	--[[
	Lavish Suramar Feast
	Item Level 110
	Use: Set out a Lavish Suramar Feast to feed up to 35 people in your raid or party!

	Restores 2400000 health and 1200000 mana over 20 sec.  Must remain seated while eating.  If you spend at least 10 seconds eating you will become well fed and gain 200 in a stat for 1 hour.
	Requires Level 101


	Hearty Feast
	Item Level 110
	Use: Set out a Hearty Feast to feed up to 35 people in your raid or party!

	Restores 2400000 health and 1200000 mana over 20 sec.  Must remain seated while eating.  If you spend at least 10 seconds eating you will become well fed and gain 150 in a stat for 1 hour.
	Requires Level 101


	Savage Feast
	Item Level 100
	Use: Feed up to 30 people in your raid or party with a Savage Feast.

	Restores 170000 health and (42500 / 5 * 20) mana over 20 sec.  If you spend at least 10 seconds eating you will become well fed and gain 100 in your attuned stat for 1 hour.
	Requires Level 91
	--]]


	-- L["[Ff]eed up to (%d+) people in your raid or party"] = true



	--[[
	Gigantic Feast
	Item Level 70
	Use: Set out a great feast that will feed your party! Restores 21252 health and (3023 / 5 * 30) mana over 30 sec. If you spend at least 10 seconds eating you will become larger!
	Requires Level 70
	--]]


	-- L["[Ff]eed your party"] = true



	--[[
	Potato Stew Feast
	Item Level 1
	Binds when picked up
	Use: Set out a feast of steak and potatoes to feed your entire raid or party!

	Restores 5% health and 5% mana per second over 20 sec.  If you spend at least 10 seconds eating you will become well fed and gain 10% Stamina for 1 hour.

	Only usable in Battlegrounds.


	Feast of Ribs
	Item Level 1
	Binds when picked up
	Use: Set out a rack of ribs to feed your entire raid or party!

	Restores 5% health and 5% mana per second over 20 sec.  If you spend at least 10 seconds eating you will become well fed and gain 10% Stamina for 1 hour.

	Only usable in Battlegrounds.


	Seafood Magnifique Feast
	Item Level 81
	Binds when picked up
	Use: Set out a seafood feast to feed your entire raid or party!

	Restores 23520 health and (3920 / 5 * 30) mana over 30 sec.  If you spend at least 10 seconds eating you will become well fed and gain 26 Stamina and 26 in another useful stat for 1 hour.
	Requires Level 80
	--]]


	-- L["[Ff]eed your entire raid or party"] = true




	if (string.find(text, L["Restores .- over 20 sec"])) then
		
		local suffix, statSuffix
		local prefix = L["Restores"]

		local captures = {}
		

		if (string.find(text, L["per second over 20 sec"])) then
			suffix = L["for 20 sec"]
			statSuffix = L["/sec"]
		else
			suffix = L["over 20 sec"]
			statSuffix = ""
		end

		for pattern,data in pairs(replacements) do
			local matches = { string.match(text, pattern) }
			local colorKey, typeKey = unpack(data)
			local color = self.colors[colorKey]

			if (#matches > 0) then
				local str = prefix .. " |c" .. color:ToAHEX() .. table.concat(matches, " - ") .. " " .. typeStringMap[typeKey] .. "|r" .. statSuffix .. " (" .. suffix .. ")" 
				table.insert(captures, str)
			end
		end

		if (#captures > 0) then
			local cd = string.match(text, L["%([%d.,]+ %a- [Cc]ooldown%)"])
	
			title = "|c" .. titleColor .. L["On Use"] .. (cd and " " .. cd or "") .. "|r\n"
			desc = table.concat(captures, "\n")
		
			newText = title .. desc .. "\n"
		end


		for i,v in ipairs(feastPatterns) do
			local feast = string.match(text, v)

			if (feast) then
				title = "\n|c" .. titleColor .. L["Feast"] .. "|r\n"
				desc = feast:gsub("^%l", string.upper) .. "."

				newText = newText .. title .. desc .. "\n"
				break
			end
		end


		local wellfed = string.match(text, L["If you spend at least 10 seconds eating you will become well fed and (.+)"])
		if (wellfed) then
			title = "\n|c" .. titleColor .. L["Well fed"] .. "|r\n"
			desc = wellfed:gsub("^%l", string.upper)

			newText = newText .. title .. desc .. "\n"
		end


	elseif (string.find(text, L["^Use: Restores"]) or string.find(text, L["^Use: Instantly restores"])) then
		local captures = {}
		local prefix = L["Restores"]
		local suffix = L["instant"]

		for pattern,data in pairs(replacements) do
			local matches = { string.match(text, pattern) }
			local colorKey, typeKey = unpack(data)
			local color = self.colors[colorKey]

			if (#matches > 0) then
				local str = prefix .. " |c" .. color:ToAHEX() .. table.concat(matches, " - ") .. " " .. typeStringMap[typeKey] .. "|r (" .. suffix .. ")" 
				table.insert(captures, str)
			end
		end

		if (#captures > 0) then
			local cd = string.match(text, L["%([%d.,]+ %a- [Cc]ooldown%)"])
	
			title = "|c" .. titleColor .. L["On Use"] .. (cd and "|r |c"  .. cdColor .. cd or "") .. "|r\n"
			desc = string.gsub(table.concat(captures, "\n"), L["%([%d.,]+ %a- [Cc]ooldown%)"], "")
		
			newText = title .. desc .. "\n"
		end
	else
		local effect = string.match(text, L["^Use: (.+)"])
		local cd = string.match(text, L["%([%d.,]+ %a- [Cc]ooldown%)"])

		if (effect) then
			title = "|c" .. titleColor .. L["On Use"] .. (cd and "|r |c"  .. cdColor .. cd or "") .. "|r\n"
			desc = string.gsub(effect, L["%([%d.,]+ %a- [Cc]ooldown%)"], "")
		
			newText = title .. desc .. "\n"
		end
	end

	if (newText) then 
		found = true
		newColor = descColorKey
	end
	-- for _,data in ipairs(self.patterns[PATTERN_USE_TRANSFORM]) do
	-- 	local pattern, colorKey = data:Unpack()
	-- 	local _, _, effect = string.find(text, pattern)

	-- 	local title, desc

	-- 	if (effect) then
	-- 		title = "|c" .. titleColor:ToAHEX() .. USE_TRANSFORMATION_MAP[pattern] .. "|r"
	-- 		desc = effect:gsub("^%l", string.upper)

	-- 		newText = title .. "\n" .. desc

	-- 		found = true
	-- 		newColor = colorKey
	-- 		break
	-- 	end
	-- end

	return found, newText, newColor
end


function ShinyStats:ParseUseSimplify(text)
	local found, newText, newColor = false

	return found, newText, newColor
end


function ShinyStats:ParseLine(text, category)
	local found, newColor = false

	for _,data in ipairs(self.patterns[category]) do
		local pattern, colorKey = data:Unpack()

		-- print("ParseLine ("..category.."):", pattern, text)

		if (text:match(pattern)) then
			found = true
			newColor = colorKey

			-- print("\t", "Match found!")
			break
		end
	end

	return found, newColor
end


function ShinyStats:ParsePartial(text, ...)		-- categories
	local found, original, newText = false, text, text
	local foundJoin = false
	local replacements = {}

	local i = 0

	for _,category in ipairs{...} do
		for _,data in ipairs(self.patterns[category]) do
			local pattern, colorKey = data:Unpack()
			local color = self.colors[colorKey]
			local colorStr

			-- print("ParsePartial ("..category.."):", pattern, text)

			if (color and color:IsEnabled()) then
				text = text:gsub(pattern, 
					function(match)
						i = i + 1

						-- print("\t", "Match found! ("..i..")")
						replacements[i] = ("|c%s%s|r"):format(color:ToAHEX(), match)

						return "{"..i.."}"
					end)
			end
		end
	end

	if (#replacements > 0) then

		-- Only run ParsePartialJoin if we have got a proper match in the parsed text earlier,
		-- effectively avoiding random "Stat-words" in names and flavor text etc.
		foundJoin, text = self:ParsePartialJoin(text)

		text = text:gsub(PARSE_TEXT_TEMPLATE_PATTERN, function(match)
				return replacements[tonumber(match)]
			end)

		found = true
		newText = text
	end

	return found, newText or text
end


function ShinyStats:ParsePartialJoin(text)
	local found, original = false, text

	for _,data in ipairs(self.patterns[PATTERN_PARTIAL_JOIN]) do
		local pattern, colorKey = data:Unpack()
		local color = self.colors[colorKey]
		local colorStr

		if (color and color:IsEnabled()) then
			text = text:gsub(pattern, 
				function(match)
					return ("|c%s%s|r"):format(color:ToAHEX(), match)
				end)
		end
	end

	if (text ~= original) then
		found = true
	end

	return found, text
end


function ShinyStats:ParseRegionText(text, itemLink, region, regionIndex, regionTable)
	local found, newText, newColor = false, false, false
	local previousRegion = regionTable[regionIndex-1]

	if (self:ShouldSkip(region)) then
		-- print("Region skipped, matching skip patterns")
		return false
	end
	--[===[@debug@
	-- print("Region Type", regionTable[regionIndex-1] and regionTable[regionIndex-1]:GetObjectType())
	--@end-debug@]===]

	if (text:match(L["^Enchant"])) then
		if (not ShinyDB:GetOption("ENABLE_ENCHANT")) then 
			--[===[@debug@
			-- print("Enchant parsing disabled!")
			--@end-debug@]===]
			return false 
		end

		found, newColor = self:ParseLine(text, PATTERN_ENCHANT)

		if (ShinyDB:GetOption("SHORTEN_ENCHANT")) then
			newText = self:Shorten(newText or text)
		end
		
		if (ShinyDB:GetOption("ENABLE_PARTIAL_MATCHES")) then
			partialFind, newText = self:ParsePartial(newText or text, PATTERN_PARTIAL)
		end

		found = found or partialFind

	elseif (text:match(L["^Equip"])) then
		if (not ShinyDB:GetOption("ENABLE_EQUIP")) then 
			--[===[@debug@
			-- print("Equip parsing disabled!")
			--@end-debug@]===]
			return false 
		end

		if (ShinyDB:GetOption("ENABLE_EQUIP_TRANSFORMATION")) then
			found, newText, newColor = self:ParseEquip(text)
		end

		if (ShinyDB:GetOption("SHORTEN_EQUIP")) then
			newText = self:Shorten(newText or text)
		end

		if (ShinyDB:GetOption("ENABLE_PARTIAL_MATCHES")) then
			partialFind, newText = self:ParsePartial(newText or text, PATTERN_PARTIAL)
		end

		found = found or partialFind

	elseif (text:match(L["^Use"])) then
		if (not ShinyDB:GetOption("ENABLE_USE")) then 
			--[===[@debug@
			-- print("Use parsing disabled!")
			--@end-debug@]===]
			return false 
		end

		found, newText, newColor = self:ParseUse(text)

	elseif (text:match(L["^Chance"])) then
		if (not ShinyDB:GetOption("ENABLE_CHANCE")) then 
			--[===[@debug@
			-- print("Chance on Hit parsing disabled!")
			--@end-debug@]===]
			return false 
		end

		found, newColor = self:ParseLine(text, PATTERN_CHANCE)

		if (ShinyDB:GetOption("SHORTEN_CHANCE")) then
			newText = self:Shorten(newText or text)
		end

	elseif (text:match(L["^%a+ Artifact Relic$"])) then
		if (not ShinyDB:GetOption("ENABLE_ARTIFACT")) then 
			--[===[@debug@
			-- print("Artifact Relic parsing disabled!")
			--@end-debug@]===]
			return false 
		end

		found, newColor = self:ParseRelic(text)

	-- elseif (previousRegion and previousRegion:GetObjectType() == "Texture") then  -- Is a socket text string
	-- 	if (not ShinyDB:GetOption("ENABLE_SOCKET_GEM")) then return false end

	--[===[ 	--[===[@debug@
	-- 	print("Found socket!", text)
	-- 	--@end-debug@]===]

	-- 	found, newColor = self:ParseLine(text, PATTERN_CATEGORY_LINE)

	-- 	if (ShinyDB:GetOption("SHORTEN_SOCKET_GEM")) then
	-- 		newText = self:Shorten(newText or text)
	-- 	end

	else
		found, newColor = self:ParseLine(text, PATTERN_LINE)

		if (ShinyDB:GetOption("SHORTEN_ENCHANT")) then
			newText = self:Shorten(newText or text)
		end

		if (ShinyDB:GetOption("ENABLE_STAT_STRIP_COLORS")) then
			newText = self:Strip(newText or text)
		end

		if (not found and ShinyDB:GetOption("ENABLE_PARTIAL_MATCHES")) then
			found, newText = self:ParsePartial(text, PATTERN_PARTIAL, PATTERN_SPECIAL)

			if (ShinyDB:GetOption("SHORTEN_SOCKET_GEM")) then
				newText = self:Shorten(newText or text)
			end
		end
	end


	return found, newText, newColor
end


function ShinyStats:ParseRegion(region, itemLink, regionIndex, regionTable)
	if (not (region and region:GetObjectType() == "FontString")) then
		return false
	end

	local original = region:GetText()
	local text = original

	if (not text or text == " " or text == "") then
		return false
	end

	if (ShinyCache:IsEnabled()) then
		local cacheResult = ShinyCache:Get(original)

		if (cacheResult) then
			--[===[@debug@
			-- print("Cached (parsed)")
			--@end-debug@]===]
			
			if (cacheResult.color and cacheResult.color:IsEnabled()) then
				-- print("\tRegion default text color set from cache")
				region:SetTextColor(cacheResult.color:ToDECA())
			end

			if (cacheResult.text) then
				-- print("\tRegion text set from cache")
				region:SetText(cacheResult.text)
			end

			return true

		elseif (cacheResult ~= nil) then
			--[===[@debug@
			-- print("Cached (ignored)")
			--@end-debug@]===]
			return false

		else
			--[===[@debug@
			-- print("Cache miss!", ShinyCache:IsEnabled(), ShinyCache:Get(original), original)
			--@end-debug@]===]
		end
	else
		--[===[@debug@
		-- print("Caching disabled!")
		--@end-debug@]===]
	end

	local found, newText, newColor = self:ParseRegionText(text, itemLink, region, regionIndex, regionTable)
	if (found) then
		--[===[@debug@
		-- print("Found")
		--@end-debug@]===]
		local color

		-- print("Found", newColor)
		if (ShinyDB:GetOption("ENABLE_STAT_DIM_USELESS") and newColor and STAT_CLASS_MAP[newColor] and not self.usefulStats[newColor]) then
			color = self.colors[SHINY_USELESS]
		else
			color = newColor and self.colors[newColor]
		end

		if (color and color:IsEnabled()) then
			region:SetTextColor(color:ToDECA())
		end

		if (newText ~= original and newText) then
			region:SetText(newText)
		end

		ShinyCache:Set(original, {
				color = color,
				text = newText ~= original and newText
			})
	else
		--[===[@debug@
		-- print("Not found", found, "-->"..(original or " ").."<--", newText, newColor)
		--@end-debug@]===]
		ShinyCache:Set(original, false)
		-- print(original, ShinyCache:Get(original))
	end

	return found, newText, newColor
end


local lastItem, lastTip, seen

function ShinyStats:ExtendTooltip(tip, itemLink)
	self:UpdateIfTainted()

	if (ShinyCache:IsTainted()) then
		ShinyCache:Clear()
	end

	local regionTable = { tip:GetRegions() }
	for regionIndex,region in ipairs(regionTable) do
		local found, newText, newColor = self:ParseRegion(region, itemLink, regionIndex, regionTable)

		if (ShinyDB:GetOption("DEBUG")) then
			if (lastItem ~= itemLink or lastTip ~= tip) then
				seen = setmetatable({}, { __mode = "kv" })
				seen[region] = true

				lastItem = itemLink
				lastTip = tip

				print(itemLink)
			elseif (not seen[region]) then
				seen[region] = true

				local text = (region and (region:GetObjectType() == "FontString")) and region:GetText()
				if (text and text ~= "" and text ~= " ") then
					local pText = newText ~= original and newText and "-->    "..newText or newColor and newColor.id or ""
					local escapedText, escapedPText = self:Escape(text),  self:Escape(pText)

					local _, relativeTo = region:GetPoint(1)
					print(region:GetPoint(1), relativeTo:GetName(), region:GetNumPoints())
					print(region:GetPoint(2), relativeTo:GetName(), region:GetNumPoints())
					print(regionIndex, region:GetName(), region:GetLeft(), string.format("%-15s%-45s    %s    %s    %s",found and "(found)" or "", text, pText, escapedText ~= escapedPText and escapedText or nil, escapedText ~= escapedPText and escapedPText or nil))
				elseif (region:GetObjectType() == "Texture" and region:GetTexture()) then
					print(regionIndex, region:GetName(), region:GetTexture())
				end

			end
		end
	end

	tip:Show()
end


function ShinyStats:HookEvents()
  if (self.eventsHooked) then return end
	self.eventsHooked = true

	self.listener:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")
	self.listener:RegisterEvent("PLAYER_TALENT_UPDATE")
end


function ShinyStats:HookTooltips()
  if (self.tooltipsHooked) then return end
  self.tooltipsHooked = true

  GameTooltip:HookScript("OnTooltipSetItem", self.TooltipHook)
  ItemRefTooltip:HookScript("OnTooltipSetItem", self.TooltipHook)

  ItemRefShoppingTooltip1:HookScript("OnTooltipSetItem", self.TooltipHook)
  ItemRefShoppingTooltip2:HookScript("OnTooltipSetItem", self.TooltipHook)

  ShoppingTooltip1:HookScript("OnTooltipSetItem", self.TooltipHook)
  ShoppingTooltip2:HookScript("OnTooltipSetItem", self.TooltipHook)


  if (WorldMapTooltip) then
 	WorldMapTooltip:HookScript("OnTooltipSetItem", self.TooltipHook)
  end


  if (WorldMapCompareTooltip1) then
 	WorldMapCompareTooltip1:HookScript("OnTooltipSetItem", self.TooltipHook)
  end


  if (AtlasLoot and AtlasLoot.db.Tooltip.tooltip == "AtlasLootTooltip" and AtlasLootTooltip) then
  	AtlasLootTooltip:HookScript("OnTooltipSetItem", self.TooltipHook)
  end


  if LinkWrangler and LinkWrangler.RegisterCallback then
    LinkWrangler.RegisterCallback("ShinyStats", function(frame,link)
      ShinyStats:ExtendTooltip(frame,link)
    end, "item")
  end

end


function ShinyStats.TooltipHook(tip)
	local _, itemLink = tip:GetItem()

	if (itemLink) then
		return ShinyStats:ExtendTooltip(tip, itemLink)
	end
end
